/* Zeo - Z/Geometry and optics computation library.
 * Copyright (C) 2005 Tomomichi Sugihara (Zhidao)
 *
 * zeo_ep - Euler parameter (unit quaternion) class.
 */

#ifndef __ZEO_EP_H__
#define __ZEO_EP_H__

#include <zeo/zeo_mat3d.h>

__BEGIN_DECLS

/* ********************************************************** */
/* CLASS: zEP
 * Euler parameter class
 * ********************************************************** */

typedef union{
  struct{
    double w;
    zVec3D v;
  } ex;
  double e[4];
} zEP;

/*! \brief create and copy Euler parameter.
 *
 * zEPCreate() creates Euler parameter \a ep.
 *
 * zEPIdent() creates Euler parameter which is equivalent to the
 * identity transformation, namely, \a ep equals to [ 1, 0, 0, 0 ].
 *
 * zEPCopy() copies Euler parameter \a src to the other \a dest.
 * \return
 * zEPCreate() and zEPIdent() return a pointer \a ep.
 * zEPCopy() returns a pointer \a dest.
 */
__EXPORT zEP *zEPCreate(zEP *ep, double theta, zVec3D *axis);
#define zEPIdent(e)  zEPCreate( e, 0, ZVEC3DZERO )
#define zEPCopy(s,d) zCopy( zEP, (s), (d) )

__EXPORT bool zEPIsIdent(zEP *ep);

/*! \brief alternate angle-axis vector, attitude matrix and Euler parameter.
 *
 * zEP2AA() converts Euler parameter \a ep to the equivalent angle-axis
 * vector \a aa.
 *
 * zAA2EP() converts an angle-axis vector \a aa to the equivalent Euler
 * parameter \a ep.
 *
 * zMat3DEP() creates a 3x3 attitude matrix \a m from Euler parameter \a ep.
 *
 * zMat3DToEP() converts a 3x3 attitude matrix \a m to the equivalent
 * Euler parameter \a ep.
 * \return
 * zAA2EP() returns a pointer \a ep.
 * zEP2AA() returns a pointer \a aa.
 * zMat3DEP() returns a pointer \a m.
 * zMat3DToEP() returns a pointer \a ep.
 */
__EXPORT zVec3D *zEP2AA(zEP *ep, zVec3D *aa);
__EXPORT zEP *zAA2EP(zVec3D *aa, zEP *ep);
__EXPORT zMat3D *zMat3DEP(zMat3D *m, zEP *ep);
__EXPORT zEP *zMat3DToEP(zMat3D *m, zEP *ep);

/*! \brief rotate a 3D vector by Euler parameter.
 *
 * zEPRotVec() rotates a 3D vector \a v by Euler parameter \a ep and puts
 * it into \a rv.
 * \return
 * zEPRotVec() returns a pointer \a rv.
 * \sa
 * zMat3DEP
 */
__EXPORT zVec3D *zEPRotVec(zEP *ep, zVec3D *v, zVec3D *rv);

/*! \brief convert rotation velocity to Euler parameter derivative.
 *
 * zEPVel2AngVel() converts the derivative of Euler parameter \a epvel
 * at the attitude represented by \a ep to the equivalent angular velocity
 * \a angvel.
 *
 * zAngVel2EPVel() converts an angular velocity \a angvel at the attitude
 * \a ep to the equivalent derivative of Euler parameter \a epvel, where
 * \a epvel is perpendicular to \a ep.
 * \return
 * zEPVel2AngVel() returns a pointer \a angvel.
 * zAngVel2EPVel() returns a pointer \a epvel.
 */
__EXPORT zVec3D *zEPVel2AngVel(zEP *epvel, zEP *ep, zVec3D *angvel);
__EXPORT zEP *zAngVel2EPVel(zVec3D *angvel, zEP *ep, zEP *epvel);

/*! \brief Euler parameter arithmetics.
 *
 * zEPRev() reverses Euler parameter \a ep1 and puts it into \a ep.
 * zEPRevDRC() directly reverses Euler parameter \a ep.
 *
 * zEPMul() multiplies Euler parameter \a ep1 by a scalar value \a k and
 * puts it into \a ep.
 * zEPMulDRC() directly multiplies Euler parameter \a ep by a scalar value \a k.
 *
 * zEPCat() multiplies Euler parameter \a ep2 by a scalar value \a k,
 * concatenates it to the other \a ep1 and puts it into \a ep.
 * zEPCatDRC() directly concatenates Euler parameter \a ep2 multiplied by
 * a scalar value \a k to the other \a ep1.
 *
 * zEPInnerProd() calculates the inner products of two Euler parameters
 * \a ep1 and \a ep2.
 * zEPNorm() calculates the norm of Euler parameter \a ep.
 *
 * zEPNormalize() directly normalizes Euler parameter \a ep.
 * \notes
 * Since Euler parameter is a class of unit quaternions, zEPNormalize() is
 * automatically called in zEPCreate() and some other functions.
 * \return
 * zEPRev(), zEPMul() and zEPCat() return a pointer \a ep.
 *
 * zEPRevDRC(), zEPMulDRC(), zEPCatDRC() and zEPNormalize() return a pointer
 * to the result Euler parameter.
 *
 * zEPInnerProd() and zEPNorm() return the result value.
 */
__EXPORT zEP *zEPSub(zEP *ep1, zEP *ep2, zEP *ep);
__EXPORT zEP *zEPRev(zEP *ep1, zEP *ep);
__EXPORT zEP *zEPMul(zEP *ep1, double k, zEP *ep);
__EXPORT zEP *zEPCat(zEP *ep1, double k, zEP *ep2, zEP *ep);
#define zEPSubDRC(e1,e)      zEPSub( e1, e, e1 )
#define zEPRevDRC(e)         zEPRev( e, e )
#define zEPMulDRC(e,k)       zEPMul( e, k, e )
#define zEPCatDRC(ep1,k,ep2) zEPCat( ep1, k, ep2, ep1 )

__EXPORT zEP *zEPDif(zEP *ep1, zEP *ep2, double dt, zEP *ep_vel);

__EXPORT double zEPInnerProd(zEP *ep1, zEP *ep2);
__EXPORT double zEPNorm(zEP *ep);
__EXPORT zEP *zEPNormalize(zEP *ep);

/*! \brief cascade a Euler parameter to another.
 */
__EXPORT zEP *zEPCascade(zEP *e1, zEP *e2, zEP *e);

/*! \brief interior division of Euler parameter.
 *
 * zEPInterDiv() calculates the interior division of two Euler parameters
 * \a ep1 and \a ep2 in accordance with the spherical interpolation.
 * \a t is the dividing ratio, which is regularly chosen within the range
 * from 0 to 1. When \a t is out of the range, zEPInterDiv() computes the
 * outer division for extrapolation. The result is put into \a ep.
 *
 * The computation is based on SLERP(spherical linear interpolation
 * proposed by K. Shoemake, 1985).
 * \return
 * zEPInterDiv() returns a pointer \a ep.
 * \sa
 * zMat3DInterDiv
 */
__EXPORT zEP *zEPInterDiv(zEP *ep1, zEP *ep2, double t, zEP *ep);

/*! \brief interior division of two attitude matrices for SLERP.
 *
 * zMat3DInterDiv() calculates the interior division of two 3x3 attitude
 * matrices \a m1 and \a m2 in accordance with the spherical interpolation.
 * \a t is the dividing ratio, which is regularly chosen within the range
 * from 0 to 1. When \a t is out of the range, zMat3DInterDiv() computes
 * the outer division for extrapolation. The result is put into \a m.
 * The computation enables SLERP(spherical linear interpolation).
 * \return
 * zMat3DInterDiv() returns a pointer \a m.
 * \sa
 * zEPInterDiv
 */
__EXPORT zMat3D *zMat3DInterDiv(zMat3D *m1, zMat3D *m2, double t, zMat3D *m);

/*! \brief output Euler parameter.
 *
 * zEPFWrite() outputs Euler parameter to the current position of a file
 * \a fp in the following format.
 *  e0 { e1, e2, e3 }
 * zEPWrite() outputs Euler parameter \a ep to the standard output.
 * \return
 * zEPFWrite() and zEPWrite() return no value.
 */
__EXPORT void zEPFWrite(FILE *fp, zEP *ep);
#define zEPWrite(e) zEPFWrite( stdout, e )

__END_DECLS

#endif /* __ZEO_EP_H__ */
