/* Zeo - Z/Geometry and optics computation library.
 * Copyright (C) 2005 Tomomichi Sugihara (Zhidao)
 *
 * zeo_color - color elements
 */

#include <zeo/zeo_color.h>
#include <math.h>

/* ********************************************************** */
/* CLASS: zRGB
 * RGB class - expression of color with RGB intensity set.
 * ********************************************************** */

static zRGB *_zRGBRatio(zRGB *rgb, char *ratio);
static zRGB *_zRGBHex(zRGB *rgb, char *hex);

/* RGB sets for black and white */
const zRGB zrgbblack = { 0.0, 0.0, 0.0 }, zrgbwhite = { 1.0, 1.0, 1.0 };

/* zRGBSet
 * - set RGB parameters.
 */
zRGB *zRGBSet(zRGB *rgb, float red, float green, float blue)
{
  rgb->r = zLimit( red,   0, 1 );
  rgb->g = zLimit( green, 0, 1 );
  rgb->b = zLimit( blue,  0, 1 );
  return rgb;
}

/* zRGBMul
 * - multiply a set of RGB parameters by another.
 */
zRGB *zRGBMul(zRGB *rgb1, zRGB *rgb2, zRGB *rgb)
{
  return zRGBSet( rgb, rgb1->r*rgb2->r, rgb1->g*rgb2->g, rgb1->b*rgb2->b );
}

/* (static)
 * _zRGBRatio
 * - decode a string of sequential floating-point values to RGB.
 */
zRGB *_zRGBRatio(zRGB *rgb, char *ratio)
{
  float r, g, b;

  sscanf( ratio, "%f:%f:%f", &r, &g, &b );
  return zRGBSet( rgb, r, g, b );
}

/* (static)
 * _zRGBHex
 * - decode a hexadecimal string to RGB.
 */
zRGB *_zRGBHex(zRGB *rgb, char *hex)
{
  int len, i;
  uint r, g, b, d;

  len = strlen( hex );
  if( len % 3 != 0 )
    ZRUNWARN( ZEO_ERR_RGB, hex, len );
  len /= 3;
  for( r=g=b=d=0, i=0; i<len; i++ ){
    r <<= 4; r |= atox_c( hex[i] );
    g <<= 4; g |= atox_c( hex[i+len] );
    b <<= 4; b |= atox_c( hex[i+len*2] );
    d <<= 4; d |= 0xf;
  }
  return zRGBSet( rgb, (float)r/d, (float)g/d, (float)b/d );
}

/* zRGBDec
 * - decode a string to RGB.
 */
zRGB *zRGBDec(zRGB *rgb, char *str)
{
  return str[0] == '#' ? _zRGBHex( rgb, str+1 ) : _zRGBRatio( rgb, str );
}

/* zRGBFRead
 * - input RGB parameter set.
 */
zRGB *zRGBFRead(FILE *fp, zRGB *rgb)
{
  float r, g, b;

  r = zFDouble(fp);
  g = zFDouble(fp);
  b = zFDouble(fp);
  return zRGBSet( rgb, r, g, b );
}

/* zRGBFWrite
 * - output RGB parameter set.
 */
void zRGBFWrite(FILE *fp, zRGB *rgb)
{
  fprintf( fp, "%.10g:%.10g:%.10g\n", rgb->r, rgb->g, rgb->b );
}

/* zRGBFWriteXML
 * - xml output.
 */
void zRGBFWriteXML(FILE *fp, zRGB *rgb)
{
  fprintf( fp, "\"%.10g %.10g %.10g\"", rgb->r, rgb->g, rgb->b );
}

/* ********************************************************** */
/* CLASS: zHSV
 * HSV class - expression of color by hue, saturation and value.
 * ********************************************************** */

/* zRGB2HSV
 * - convert RGB to HSV.
 */
zHSV *zRGB2HSV(zRGB *rgb, zHSV *hsv)
{
  float max, min, d, phase, phase0;

  /* find principal primary color */
  if( rgb->r > rgb->g ){
    if( rgb->r > rgb->b ){
      max = rgb->r;
      min = zMin( rgb->g, rgb->b );
      phase = rgb->g - rgb->b;
      phase0 = 0;
    } else{
      max = rgb->b;
      min = rgb->g;
      phase = rgb->r - rgb->g;
      phase0 = 240;
    }
  } else{
    if( rgb->g > rgb->b ){
      max = rgb->g;
      min = zMin( rgb->r, rgb->b );
      phase = rgb->b - rgb->r;
      phase0 = 120;
    } else{
      max = rgb->b;
      min = rgb->r;
      phase = rgb->r - rgb->g;
      phase0 = 240;
    }
  }
  d = max - min;
  /* conversion */
  hsv->hue =   d == 0 ? 0 : 60.0 * phase / d + phase0;
  if( hsv->hue < 0 ) hsv->hue += 360;
  hsv->sat = max == 0 ? 0 : d / max;
  hsv->val = max;
  return hsv;
}

/* zHSV2RGB
 * - convert HSV to RGB.
 */
zRGB *zHSV2RGB(zHSV *hsv, zRGB *rgb)
{
  float h, f, p, q, t;
  int i;

  h = hsv->hue / 60;
  i = (int)floor( h ) % 6;
  f = h - i;
  p = hsv->val * ( 1 - hsv->sat );
  q = hsv->val * ( 1 - f * hsv->sat );
  t = hsv->val * ( 1 - (1-f) * hsv->sat );
  switch( i ){
  case 0: zRGBSet( rgb, hsv->val, t, p ); break;
  case 1: zRGBSet( rgb, q, hsv->val, p ); break;
  case 2: zRGBSet( rgb, p, hsv->val, t ); break;
  case 3: zRGBSet( rgb, p, q, hsv->val ); break;
  case 4: zRGBSet( rgb, t, p, hsv->val ); break;
  case 5: zRGBSet( rgb, hsv->val, p, q ); break;
  default: ;
  }
  return rgb;
}

/* zHSVFWrite
 * - output HSV parameter set.
 */
void zHSVFWrite(FILE *fp, zHSV *hsv)
{
  fprintf( fp, "%.10g:%.10g:%.10g\n", hsv->hue, hsv->sat, hsv->val );
}
